from litestar.cli.main import litestar_group  # pragma: no cover


def run_cli() -> None:  # pragma: no cover
    """Application Entrypoint."""
    litestar_group()


if __name__ == "__main__":  # pragma: no cover
    run_cli()
